/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

// Arbres binaires de recherche, mutables

#include "prelude.h"
#include "bst.h"

// 1. on commence par les noeuds

typedef struct Node {
  struct Node *left;
  char *key;
  int value;
  struct Node *right;
} node;

node *bst_createn(char *k, int v) {
  node *t = malloc(sizeof(struct Node));
  t->key = k;
  t->value = v;
  t->left = t->right = NULL;
  return t;
}

bool bst_containsn(node *t, char *k) {
  if (t == NULL) return false;
  int c = strcmp(k, t->key);
  if (c < 0) return bst_containsn(t->left, k);
  if (c > 0) return bst_containsn(t->right, k);
  return true;
}

// recherche avec une fonction récursive ...
int bst_getn_(node *t, char *k) {
  if (t == NULL) return -1;
  int c = strcmp(k, t->key);
  if (c < 0) return bst_getn_(t->left, k);
  if (c > 0) return bst_getn_(t->right, k);
  return t->value;
}
// ... ou avec une boucle
int bst_getn(node *t, char *k) {
  while (t != NULL) {
    int c = strcmp(k, t->key);
    if (c == 0) return t->value;
    t = c < 0 ? t->left : t->right;
  }
  return -1;
}

node *bst_putn(node *t, char *k, int v) {
  if (t == NULL) return bst_createn(k, v);
  int c = strcmp(k, t->key);
  if      (c < 0) { t->left  = bst_putn(t->left, k, v); }
  else if (c > 0) { t->right = bst_putn(t->right, k, v); }
  else t->value = v;
  return t;
}

void bst_deleten(node *t) {
  if (t == NULL) return;
  bst_deleten(t->left);
  bst_deleten(t->right);
  free(t);
}

void bst_printn(node *t) {
  if (t == NULL) return;
  bst_printn(t->left);
  printf("%s -> %d; ", t->key, t->value);
  bst_printn(t->right);
}

char *bst_min_eltn(node *t) {
  if (t == NULL) return NULL;
  while (t->left != NULL)
    t = t->left;
  return t->key;
}

node *bst_rotate_right(node *t) {
  if (t == NULL || t->left == NULL) return t;
  node *l = t->left;
  t->left = l->right;
  l->right = t;
  return l;
}
node *bst_rotate_left(node *t) {
  if (t == NULL || t->right == NULL) return t;
  node *r = t->right;
  t->right = r->left;
  r->left = t;
  return r;
}

// 2. puis on encapsule l'arbre dans une structure

struct Bst {
  node *root;
};

bst *bst_create(void) {
  bst *t = malloc(sizeof(struct Bst));
  t->root = NULL;
  return t;
}

bool bst_is_empty(bst *t) {
  return t->root == NULL;
}

void bst_put(bst *t, char *k, int v) {
  t->root = bst_putn(t->root, k, v);
}

bool bst_contains(bst *t, char *k) {
  return bst_containsn(t->root, k);
}

int bst_get(bst *t, char *k) {
  return bst_getn(t->root, k);
}

char *bst_min_elt(bst *t) {
  return bst_min_eltn(t->root);
}

void bst_print(bst *t) {
  printf("[");
  bst_printn(t->root);
  printf("]\n");
}

void bst_delete(bst *t) {
  bst_deleten(t->root);
  free(t);
}

